"""causal_curve module"""

import warnings

from statsmodels.genmod.generalized_linear_model import DomainWarning

from causal_curve.gps_classifier import GPS_Classifier
from causal_curve.gps_regressor import GPS_Regressor

from causal_curve.tmle_regressor import TMLE_Regressor
from causal_curve.mediation import Mediation


# Suppress statsmodel warning for gamma family GLM
warnings.filterwarnings("ignore", category=DomainWarning)
warnings.filterwarnings("ignore", category=UserWarning)



"""
Defines the Generalized Prospensity Score (GPS) regressor model class
"""

import numpy as np

from causal_curve.gps_core import GPS_Core


class GPS_Regressor(GPS_Core):
    """
    A GPS tool that handles continuous outcomes. Inherits the GPS_core
    base class. See that base class code its docstring for more details.
    """

    def __init__(
        self,
        gps_family=None,
        treatment_grid_num=100,
        lower_grid_constraint=0.01,
        upper_grid_constraint=0.99,
        spline_order=3,
        n_splines=30,
        lambda_=0.5,
        max_iter=100,
        random_seed=None,
        verbose=False,
    ):
        GPS_Core.__init__(
            self,
            gps_family=None,
            treatment_grid_num=100,
            lower_grid_constraint=0.01,
            upper_grid_constraint=0.99,
            spline_order=3,
            n_splines=30,
            lambda_=0.5,
            max_iter=100,
            random_seed=None,
            verbose=False,
        )

    def _cdrc_predictions_continuous(self, ci):
        """Returns the predictions of CDRC for each value of the treatment grid. Essentially,
        we're making predictions using the original treatment and gps_at_grid.
        To be used when the outcome of interest is continuous.
        """

        # To keep track of cdrc predictions, we create an empty 3d array of shape
        # (n_samples, treatment_grid_num, 3). The last dimension is of length 3 because
        # we are going to keep track of the point estimate of the prediction, as well as
        # the lower and upper bounds of the prediction interval
        cdrc_preds = np.zeros((len(self.T), self.treatment_grid_num, 3), dtype=float)

        # Loop through each of the grid values, predict point estimate and get prediction interval
        for i in range(0, self.treatment_grid_num):
            temp_T = np.repeat(self.grid_values[i], repeats=len(self.T))
            temp_gps = self.gps_at_grid[:, i]
            temp_cdrc_preds = self.gam_results.predict(
                np.column_stack((temp_T, temp_gps))
            )
            temp_cdrc_interval = self.gam_results.confidence_intervals(
                np.column_stack((temp_T, temp_gps)), width=ci
            )
            temp_cdrc_lower_bound = temp_cdrc_interval[:, 0]
            temp_cdrc_upper_bound = temp_cdrc_interval[:, 1]
            cdrc_preds[:, i, 0] = temp_cdrc_preds
            cdrc_preds[:, i, 1] = temp_cdrc_lower_bound
            cdrc_preds[:, i, 2] = temp_cdrc_upper_bound

        return np.round(cdrc_preds, 3)

    def point_estimate(self, T):
        """Calculates point estimate within the CDRC given treatment values.
        Can only be used when outcome is continuous. Can be estimate for a single
        data point or can be run in batch for many observations. Extrapolation will produce
        untrustworthy results; the provided treatment should be within
        the range of the training data.

        Parameters
        ----------
        T: Numpy array, shape (n_samples,)
            A continuous treatment variable.

        Returns
        ----------
        array: Numpy array
            Contains a set of CDRC point estimates

        """
        if self.outcome_type != "continuous":
            raise TypeError("Your outcome must be continuous to use this function!")

        return np.apply_along_axis(self._create_point_estimate, 0, T.reshape(1, -1))

    def _create_point_estimate(self, T):
        """Takes a single treatment value and produces a single point estimate
        in the case of a continuous outcome.
        """
        return self.gam_results.predict(
            np.array([T, self.gps_function(T).mean()]).reshape(1, -1)
        )

    def point_estimate_interval(self, T, ci=0.95):
        """Calculates the prediction confidence interval associated with a point estimate
        within the CDRC given treatment values. Can only be used
        when outcome is continuous. Can be estimate for a single data point
        or can be run in batch for many observations. Extrapolation will produce
        untrustworthy results; the provided treatment should be within
        the range of the training data.

        Parameters
        ----------
        T: Numpy array, shape (n_samples,)
            A continuous treatment variable.
        ci: float (default = 0.95)
            The desired confidence interval to produce. Default value is 0.95, corresponding
            to 95% confidence intervals. bounded (0, 1.0).

        Returns
        ----------
        array: Numpy array
            Contains a set of CDRC prediction intervals ([lower bound, higher bound])

        """
        if self.outcome_type != "continuous":
            raise TypeError("Your outcome must be continuous to use this function!")

        return np.apply_along_axis(
            self._create_point_estimate_interval, 0, T.reshape(1, -1), width=ci
        ).T.reshape(-1, 2)

    def _create_point_estimate_interval(self, T, width):
        """Takes a single treatment value and produces confidence interval
        associated with a point estimate in the case of a continuous outcome.
        """
        return self.gam_results.prediction_intervals(
            np.array([T, self.gps_function(T).mean()]).reshape(1, -1), width=width
        )



"""
Core classes (with basic methods) that will be invoked when other, model classes are defined
"""

import numpy as np
from scipy.stats import norm


class Core:
    """Base class for causal_curve module"""

    def __init__(self):
        __version__ = "1.0.1"

    def get_params(self):
        """Returns a dict of all of the object's user-facing parameters

        Parameters
        ----------
        None

        Returns
        -------
        self: object
        """
        attrs = self.__dict__
        return dict(
            [(k, v) for k, v in list(attrs.items()) if (k[0] != "_") and (k[-1] != "_")]
        )

    def if_verbose_print(self, string):
        """Prints the input statement if verbose is set to True

        Parameters
        ----------
        string: str, some string to be printed

        Returns
        ----------
        None

        """
        if self.verbose:
            print(string)

    @staticmethod
    def rand_seed_wrapper(random_seed=None):
        """Sets the random seed using numpy

        Parameters
        ----------
        random_seed: int, random seed number

        Returns
        ----------
        None
        """
        if random_seed is None:
            pass
        else:
            np.random.seed(random_seed)

    @staticmethod
    def calculate_z_score(ci):
        """Calculates the critical z-score for a desired two-sided,
        confidence interval width.

        Parameters
        ----------
        ci: float, the confidence interval width (e.g. 0.95)

        Returns
        -------
        Float, critical z-score value
        """
        return norm.ppf((1 + ci) / 2)

    @staticmethod
    def clip_negatives(number):
        """Helper function to clip negative numbers to zero

        Parameters
        ----------
        number: int or float, any number that needs a floor at zero

        Returns
        -------
        Int or float of modified value

        """
        if number < 0:
            return 0
        return number



"""
Defines the Generalized Prospensity Score (GPS) Core model class
"""

import contextlib
import io
from pprint import pprint

import numpy as np
import pandas as pd
from pandas.api.types import is_float_dtype, is_integer_dtype, is_numeric_dtype
from pygam import LinearGAM, LogisticGAM, s
from scipy.special import logit
from scipy.stats import gamma, norm
import statsmodels.api as sm
from statsmodels.genmod.families.links import inverse_power as Inverse_Power
from statsmodels.tools.tools import add_constant

from causal_curve.core import Core


class GPS_Core(Core):
    """
    In a multi-stage approach, this computes the generalized propensity score (GPS) function,
    and uses this in a generalized additive model (GAM) to correct treatment prediction of
    the outcome variable. Assumes continuous treatment, but the outcome variable may be
    continuous or binary.

    WARNING:

    -This algorithm assumes you've already performed the necessary transformations to
    categorical covariates (i.e. these variables are already one-hot encoded and
    one of the categories is excluded for each set of dummy variables).

    -Please take care to ensure that the "ignorability" assumption is met (i.e.
    all strong confounders are captured in your covariates and there is no
    informative censoring), otherwise your results will be biased, sometimes strongly so.

    Parameters
    ----------

    gps_family: str, optional (default = None)
        Is used to determine the family of the glm used to model the GPS function.
        Look at the distribution of your treatment variable to determine which
        family is more appropriate.
        Possible values:

        - 'normal'
        - 'lognormal'
        - 'gamma'
        - None : (best-fitting family automatically chosen)

    treatment_grid_num: int, optional (default = 100)
        Takes the treatment, and creates a quantile-based grid across its values.
        For instance, if the number 6 is selected, this means the algorithm will only take
        the 6 treatment variable values at approximately the 0, 20, 40, 60, 80, and 100th
        percentiles to estimate the causal dose response curve.
        Higher value here means the final curve will be more finely estimated,
        but also increases computation time. Default is usually a reasonable number.

    lower_grid_constraint:  float, optional(default = 0.01)
        This adds an optional constraint of the lower side of the treatment grid.
        Sometimes data near the minimum values of the treatment are few in number
        and thus generate unstable estimates. By default, this clips the bottom 1 percentile
        or lower of treatment values. This can be as low as 0, indicating there is no
        lower limit to how much treatment data is considered.

    upper_grid_constraint: float, optional (default = 0.99)
        See above parameter. Just like above, but this is an upper constraint.
        By default, this clips the top 99th percentile or higher of treatment values.
        This can be as high as 1.0, indicating there is no upper limit to how much
        treatment data is considered.

    spline_order: int, optional (default = 3)
        Order of the splines to use fitting the final GAM.
        Must be integer >= 1. Default value creates cubic splines.

    n_splines: int, optional (default = 30)
        Number of splines to use for the treatment and GPS in the final GAM.
        Must be integer >= 2. Must be non-negative.

    lambda_: int or float, optional (default = 0.5)
        Strength of smoothing penalty. Must be a positive float.
        Larger values enforce stronger smoothing.

    max_iter: int, optional (default = 100)
        Maximum number of iterations allowed for the maximum likelihood algo to converge.

    random_seed: int, optional (default = None)
        Sets the random seed.

    verbose: bool, optional (default = False)
        Determines whether the user will get verbose status updates.


    Attributes
    ----------

    grid_values: array of shape (treatment_grid_num, )
        The gridded values of the treatment variable. Equally spaced.

    best_gps_family: str
        If no gps_family is specified and the algorithm chooses the best glm family, this is
        the name of the family that was chosen.

    gps_deviance: float
        The GPS model deviance

    gps: array of shape (number of observations, )
        The calculated GPS for each observation

    gam_results: `pygam.LinearGAM` class
        trained model of `LinearGAM` class, from pyGAM library


    Methods
    ----------
    fit: (self, T, X, y)
        Fits the causal dose-response model.

    calculate_CDRC: (self, ci)
        Calculates the CDRC (and confidence interval) from trained model.

    predict: (self, T)
        Calculates point estimate within the CDRC given treatment values.
        Can only be used when outcome is continuous.

    predict_interval: (self, T, ci)
        Calculates the prediction confidence interval associated with a point estimate
        within the CDRC given treatment values. Can only be used when outcome is continuous.

    predict_log_odds: (self, T)
        Calculates the predicted log odds of the highest integer class. Can
        only be used when the outcome is binary.

    print_gam_summary: (self)
        Prints pyGAM text summary of GAM predicting outcome from the treatment and the GPS.


    Examples
    --------
    >>> from causal_curve import GPS
    >>> gps = GPS(treatment_grid_num = 200, random_seed = 512)
    >>> gps.fit(T = df['Treatment'], X = df[['X_1', 'X_2']], y = df['Outcome'])
    >>> gps_results = gps.calculate_CDRC(0.95)
    >>> treatment_points = np.array([10,15,20,25])
    >>> preds = gps.predict(treatment_points)
    >>> conf_ints = gps.predict_interval(treatment_points, 0.95)


    References
    ----------

    Galagate, D. Causal Inference with a Continuous Treatment and Outcome: Alternative
    Estimators for Parametric Dose-Response function with Applications. PhD thesis, 2016.

    Moodie E and Stephens DA. Estimation of dose–response functions for
    longitudinal data using the generalised propensity score. In: Statistical Methods in
    Medical Research 21(2), 2010, pp.149–166.

    Hirano K and Imbens GW. The propensity score with continuous treatments.
    In: Gelman A and Meng XL (eds) Applied bayesian modeling and causal inference
    from incomplete-data perspectives. Oxford, UK: Wiley, 2004, pp.73–84.
    """

    def __init__(
        self,
        gps_family=None,
        treatment_grid_num=100,
        lower_grid_constraint=0.01,
        upper_grid_constraint=0.99,
        spline_order=3,
        n_splines=30,
        lambda_=0.5,
        max_iter=100,
        random_seed=None,
        verbose=False,
    ):

        self.gps_family = gps_family
        self.treatment_grid_num = treatment_grid_num
        self.lower_grid_constraint = lower_grid_constraint
        self.upper_grid_constraint = upper_grid_constraint
        self.spline_order = spline_order
        self.n_splines = n_splines
        self.lambda_ = lambda_
        self.max_iter = max_iter
        self.random_seed = random_seed
        self.verbose = verbose

        # Validate the params
        self._validate_init_params()
        self.rand_seed_wrapper()

        self.if_verbose_print("Using the following params for GPS model:")
        if self.verbose:
            pprint(self.get_params(), indent=4)

    def _validate_init_params(self):
        """
        Checks that the params used when instantiating GPS model are formatted correctly
        """
        # Checks for gps_family param
        if not isinstance(self.gps_family, (str, type(None))):
            raise TypeError(
                f"gps_family parameter must be a string or None "
                f"but found type {type(self.gps_family)}"
            )

        if (isinstance(self.gps_family, str)) and (
            self.gps_family not in ["normal", "lognormal", "gamma"]
        ):
            raise ValueError(
                f"gps_family parameter must take on values of "
                f"'normal', 'lognormal', or 'gamma', but found {self.gps_family}"
            )

        # Checks for treatment_grid_num
        if not isinstance(self.treatment_grid_num, int):
            raise TypeError(
                f"treatment_grid_num parameter must be an integer, "
                f"but found type {type(self.treatment_grid_num)}"
            )

        if (isinstance(self.treatment_grid_num, int)) and self.treatment_grid_num < 10:
            raise ValueError(
                f"treatment_grid_num parameter should be >= 10 so your final curve "
                f"has enough resolution, but found value {self.treatment_grid_num}"
            )

        if (
            isinstance(self.treatment_grid_num, int)
        ) and self.treatment_grid_num >= 1000:
            raise ValueError("treatment_grid_num parameter is too high!")

        # Checks for lower_grid_constraint
        if not isinstance(self.lower_grid_constraint, float):
            raise TypeError(
                f"lower_grid_constraint parameter must be a float, "
                f"but found type {type(self.lower_grid_constraint)}"
            )

        if (
            isinstance(self.lower_grid_constraint, float)
        ) and self.lower_grid_constraint < 0:
            raise ValueError(
                f"lower_grid_constraint parameter cannot be < 0, "
                f"but found value {self.lower_grid_constraint}"
            )

        if (
            isinstance(self.lower_grid_constraint, float)
        ) and self.lower_grid_constraint >= 1.0:
            raise ValueError(
                f"lower_grid_constraint parameter cannot >= 1.0, "
                f"but found value {self.lower_grid_constraint}"
            )

        # Checks for upper_grid_constraint
        if not isinstance(self.upper_grid_constraint, float):
            raise TypeError(
                f"upper_grid_constraint parameter must be a float, "
                f"but found type {type(self.upper_grid_constraint)}"
            )

        if (
            isinstance(self.upper_grid_constraint, float)
        ) and self.upper_grid_constraint <= 0:
            raise ValueError(
                f"upper_grid_constraint parameter cannot be <= 0, "
                f"but found value {self.upper_grid_constraint}"
            )

        if (
            isinstance(self.upper_grid_constraint, float)
        ) and self.upper_grid_constraint > 1.0:
            raise ValueError(
                f"upper_grid_constraint parameter cannot > 1.0, "
                f"but found value {self.upper_grid_constraint}"
            )

        # Checks for lower_grid_constraint isn't higher than upper_grid_constraint
        if self.lower_grid_constraint >= self.upper_grid_constraint:
            raise ValueError(
                "lower_grid_constraint should be lower than upper_grid_constraint!"
            )

        # Checks for spline_order
        if not isinstance(self.spline_order, int):
            raise TypeError(
                f"spline_order parameter must be an integer, "
                f"but found type {type(self.spline_order)}"
            )

        if (isinstance(self.spline_order, int)) and self.spline_order < 1:
            raise ValueError(
                f"spline_order parameter should be >= 1, but found {self.spline_order}"
            )

        if (isinstance(self.spline_order, int)) and self.spline_order >= 30:
            raise ValueError("spline_order parameter is too high!")

        # Checks for n_splines
        if not isinstance(self.n_splines, int):
            raise TypeError(
                f"n_splines parameter must be an integer, but found type {type(self.n_splines)}"
            )

        if (isinstance(self.n_splines, int)) and self.n_splines < 2:
            raise ValueError(
                f"n_splines parameter should be >= 2, but found {self.n_splines}"
            )

        if (isinstance(self.n_splines, int)) and self.n_splines >= 100:
            raise ValueError("n_splines parameter is too high!")

        # Checks for lambda_
        if not isinstance(self.lambda_, (int, float)):
            raise TypeError(
                f"lambda_ parameter must be an int or float, but found type {type(self.lambda_)}"
            )

        if (isinstance(self.lambda_, (int, float))) and self.lambda_ <= 0:
            raise ValueError(
                f"lambda_ parameter should be > 0, but found {self.lambda_}"
            )

        if (isinstance(self.lambda_, (int, float))) and self.lambda_ >= 1000:
            raise ValueError("lambda_ parameter is too high!")

        # Checks for max_iter
        if not isinstance(self.max_iter, int):
            raise TypeError(
                f"max_iter parameter must be an int, but found type {type(self.max_iter)}"
            )

        if (isinstance(self.max_iter, int)) and self.max_iter <= 10:
            raise ValueError(
                "max_iter parameter is too low! Results won't be reliable!"
            )

        if (isinstance(self.max_iter, int)) and self.max_iter >= 1e6:
            raise ValueError("max_iter parameter is unnecessarily high!")

        # Checks for random_seed
        if not isinstance(self.random_seed, (int, type(None))):
            raise TypeError(
                f"random_seed parameter must be an int, but found type {type(self.random_seed)}"
            )

        if (isinstance(self.random_seed, int)) and self.random_seed < 0:
            raise ValueError("random_seed parameter must be > 0")

        # Checks for verbose
        if not isinstance(self.verbose, bool):
            raise TypeError(
                f"verbose parameter must be a boolean type, but found type {type(self.verbose)}"
            )

    def _validate_fit_data(self):
        """Verifies that T, X, and y are formatted the right way"""
        # Checks for T column
        if not is_float_dtype(self.T):
            raise TypeError("Treatment data must be of type float")

        # Make sure all X columns are float or int
        if isinstance(self.X, pd.Series):
            if not is_numeric_dtype(self.X):
                raise TypeError(
                    "All covariate (X) columns must be int or float type (i.e. must be numeric)"
                )

        elif isinstance(self.X, pd.DataFrame):
            for column in self.X:
                if not is_numeric_dtype(self.X[column]):
                    raise TypeError(
                        "All covariate (X) columns must be int or float type "
                        "(i.e. must be numeric)"
                    )

        # Checks for Y column
        if not (is_float_dtype(self.y) or is_integer_dtype(self.y)):
            raise TypeError("Outcome data must be of type float or integer")

        if is_integer_dtype(self.y) and (
            not np.array_equal(np.sort(self.y.unique()), np.array([0, 1]))
        ):
            raise TypeError(
                "If your outcome data is of type integer (binary outcome),"
                "it should only contain 1's and 0's."
            )

    def _grid_values(self):
        """Produces initial grid values for the treatment variable"""
        return np.quantile(
            self.T,
            q=np.linspace(
                start=self.lower_grid_constraint,
                stop=self.upper_grid_constraint,
                num=self.treatment_grid_num,
            ),
        )

    def fit(self, T, X, y):
        """Fits the GPS causal dose-response model. For now, this only accepts pandas columns.
        While the treatment variable must be continuous (or ordinal with many levels), the
        outcome variable may be continuous or binary. You *must* provide
        at least one covariate column.

        Parameters
        ----------
        T: array-like, shape (n_samples,)
            A continuous treatment variable.
        X: array-like, shape (n_samples, m_features)
            Covariates, where n_samples is the number of samples
            and m_features is the number of features. Features can be a mix of continuous
            and nominal/categorical variables.
        y: array-like, shape (n_samples,)
            Outcome variable. May be continuous or binary. If continuous, this must
            be a series of type `float`, if binary must be a series of type `integer`.

        Returns
        ----------
        self : object

        """
        self.T = T.reset_index(drop=True, inplace=False)
        self.X = X.reset_index(drop=True, inplace=False)
        self.y = y.reset_index(drop=True, inplace=False)

        # Determine what type of outcome variable we're working with
        if is_float_dtype(self.y):
            self.outcome_type = "continuous"
        elif is_integer_dtype(self.y):
            self.outcome_type = "binary"

        self.if_verbose_print(
            f"Determined the outcome variable is of type {self.outcome_type}..."
        )

        # Validate this input data
        self._validate_fit_data()

        # Create grid_values
        self.grid_values = self._grid_values()

        # Determine which GPS family to use
        self._determine_gps_function()

        # Estimate the GPS
        self.if_verbose_print("Saving GPS values...")

        self.gps = self.gps_function(self.T)

        # Create GAM that predicts outcome from the treatment and GPS
        self.if_verbose_print("Fitting GAM using treatment and GPS...")

        # Save model results
        self.gam_results = self._fit_gam()

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            self.gam_results.summary()

        self._gam_summary_str = f.getvalue()

        self.if_verbose_print(
            "Calculating many CDRC estimates for each treatment grid value..."
        )

        # Loop over all grid values (`treatment_grid_num` in total)
        # and give GPS loading for each observation in the dataset
        self.gps_at_grid = self._gps_values_at_grid()

    def calculate_CDRC(self, ci=0.95):
        """Using the results of the fitted model, this generates a dataframe of
        point estimates for the CDRC at each of the values of the
        treatment grid. Connecting these estimates will produce the overall
        estimated CDRC. Confidence interval is returned as well.

        Parameters
        ----------
        ci: float (default = 0.95)
            The desired confidence interval to produce. Default value is 0.95, corresponding
            to 95% confidence intervals. bounded (0, 1.0).

        Returns
        ----------
        dataframe: Pandas dataframe
            Contains treatment grid values, the CDRC point estimate at that value,
            and the associated lower and upper confidence interval bounds at that point.

        self: object

        """
        self._validate_calculate_CDRC_params(ci)

        self.if_verbose_print(
            """
            Generating predictions for each value of treatment grid,
            and averaging to get the CDRC..."""
        )

        # Create CDRC predictions from trained GAM
        # If working with a continuous outcome variable, use this path
        if self.outcome_type == "continuous":
            self._cdrc_preds = self._cdrc_predictions_continuous(ci)

            results = []

            for i in range(0, self.treatment_grid_num):
                temp_grid_value = self.grid_values[i]
                temp_point_estimate = self._cdrc_preds[:, i, 0].mean()
                mean_ci_width = (
                    self._cdrc_preds[:, i, 2].mean() - self._cdrc_preds[:, i, 1].mean()
                ) / 2
                temp_lower_bound = temp_point_estimate - mean_ci_width
                temp_upper_bound = temp_point_estimate + mean_ci_width
                results.append(
                    [
                        temp_grid_value,
                        temp_point_estimate,
                        temp_lower_bound,
                        temp_upper_bound,
                    ]
                )

            outcome_name = "Causal_Dose_Response"

        # If working with a binary outcome variable, use this path
        else:
            self._cdrc_preds = self._cdrc_predictions_binary(ci)

            # Capture the first prediction's mean log odds.
            # This will serve as a reference for calculating the odds ratios
            log_odds_reference = self._cdrc_preds[:, 0, 0].mean()

            results = []

            for i in range(0, self.treatment_grid_num):
                temp_grid_value = self.grid_values[i]

                temp_log_odds_estimate = (
                    self._cdrc_preds[:, i, 0].mean() - log_odds_reference
                )
                temp_OR_estimate = np.exp(temp_log_odds_estimate)

                temp_lower_bound = np.exp(
                    temp_log_odds_estimate
                    - (self.calculate_z_score(ci) * self._cdrc_preds[:, i, 1].mean())
                )
                temp_upper_bound = np.exp(
                    temp_log_odds_estimate
                    + (self.calculate_z_score(ci) * self._cdrc_preds[:, i, 1].mean())
                )
                results.append(
                    [
                        temp_grid_value,
                        temp_OR_estimate,
                        temp_lower_bound,
                        temp_upper_bound,
                    ]
                )

            outcome_name = "Causal_Odds_Ratio"

        return pd.DataFrame(
            results, columns=["Treatment", outcome_name, "Lower_CI", "Upper_CI"]
        ).round(3)

    @staticmethod
    def _validate_calculate_CDRC_params(ci):
        """Validates the parameters given to `calculate_CDRC`"""

        if not isinstance(ci, float):
            raise TypeError(
                f"`ci` parameter must be an float, but found type {type(ci)}"
            )

        if isinstance(ci, float) and ((ci <= 0) or (ci >= 1.0)):
            raise ValueError("`ci` parameter should be between (0, 1)")

    def _gps_values_at_grid(self):
        """Returns an array where we get the GPS-derived values for each element
        of the treatment grid. Resulting array will be of shape (n_samples, treatment_grid_num)
        """
        # Creates an empty 2d array of shape (n_samples, treatment_grid_num)
        gps_at_grid = np.zeros((len(self.T), self.treatment_grid_num), dtype=float)

        # Loop over all grid values
        for i in range(0, self.treatment_grid_num):
            gps_at_grid[:, i] = self.gps_function(self.grid_values[i])

        return gps_at_grid

    def print_gam_summary(self):
        """Prints the GAM model summary (uses pyGAM's output)

        Parameters
        ----------
        None

        Returns
        ----------
        self: object
        """
        print(self._gam_summary_str)

    def _fit_gam(self):
        """Fits a GAM that predicts the outcome (continuous or binary) from the treatment and GPS"""

        X = np.column_stack((self.T.values, self.gps))
        y = np.asarray(self.y)

        model_type_dict = {"continuous": LinearGAM, "binary": LogisticGAM}

        return model_type_dict[self.outcome_type](
            s(0, n_splines=self.n_splines, spline_order=self.spline_order)
            + s(1, n_splines=self.n_splines, spline_order=self.spline_order),
            max_iter=self.max_iter,
            lam=self.lambda_,
        ).fit(X, y)

    def _determine_gps_function(self):
        """Based on the user input, distribution of treatment values, and/or model deviances,
        this function determines which GPS function family should be used.
        """

        # If any negative values in treatment, you must use the normal GLM family.
        if any(self.T <= 0):
            self.best_gps_family = "normal"
            self.gps_function, self.gps_deviance = self._create_normal_gps_function()
            self.if_verbose_print(
                """Must fit `normal` GLM family to model treatment since
                treatment takes on zero or negative values..."""
            )

        # If treatment has no negative values and user provides in put, use that.
        elif (all(self.T > 0)) & (not isinstance(self.gps_family, type(None))):
            self.if_verbose_print(f"Fitting GPS model of family '{self.gps_family}'...")

            if self.gps_family == "normal":
                self.best_gps_family = "normal"
                (
                    self.gps_function,
                    self.gps_deviance,
                ) = self._create_normal_gps_function()
            elif self.gps_family == "lognormal":
                self.best_gps_family = "lognormal"
                (
                    self.gps_function,
                    self.gps_deviance,
                ) = self._create_lognormal_gps_function()
            elif self.gps_family == "gamma":
                self.best_gps_family = "gamma"
                self.gps_function, self.gps_deviance = self._create_gamma_gps_function()

        # If no zero or negative treatment values and user didn't provide
        # input, figure out best-fitting family
        elif (all(self.T > 0)) & (isinstance(self.gps_family, type(None))):
            self.if_verbose_print(
                "Fitting several GPS models and" " picking the best fitting one..."
            )

            (
                self.best_gps_family,
                self.gps_function,
                self.gps_deviance,
            ) = self._find_best_gps_model()

            self.if_verbose_print(
                f"""Best fitting model was {self.best_gps_family}, which
                    produced a deviance of {self.gps_deviance}"""
            )

    def _create_normal_gps_function(self):
        """Models the GPS using a GLM of the Gaussian family"""
        normal_gps_model = sm.GLM(
            self.T, add_constant(self.X), family=sm.families.Gaussian()
        ).fit()

        pred_treat = normal_gps_model.fittedvalues
        sigma = np.std(normal_gps_model.resid_response)

        def gps_function(treatment_val, pred_treat=pred_treat, sigma=sigma):
            return norm.pdf(treatment_val, pred_treat, sigma)

        return gps_function, normal_gps_model.deviance

    def _create_lognormal_gps_function(self):
        """Models the GPS using a GLM of the Gaussian family (assumes treatment is lognormal)"""
        lognormal_gps_model = sm.GLM(
            np.log(self.T), add_constant(self.X), family=sm.families.Gaussian()
        ).fit()

        pred_log_treat = lognormal_gps_model.fittedvalues
        sigma = np.std(lognormal_gps_model.resid_response)

        def gps_function(treatment_val, pred_log_treat=pred_log_treat, sigma=sigma):
            return norm.pdf(np.log(treatment_val), pred_log_treat, sigma)

        return gps_function, lognormal_gps_model.deviance

    def _create_gamma_gps_function(self):
        """Models the GPS using a GLM of the Gamma family"""
        gamma_gps_model = sm.GLM(
            self.T, add_constant(self.X), family=sm.families.Gamma(Inverse_Power())
        ).fit()

        mu = gamma_gps_model.mu
        scale = gamma_gps_model.scale
        shape = mu / gamma_gps_model.scale

        def gps_function(treatment_val):
            return gamma.pdf(treatment_val, a=shape, loc=0, scale=scale)

        return gps_function, gamma_gps_model.deviance

    def _find_best_gps_model(self):
        """If user doesn't provide a GLM family for modeling the GPS, this function compares
        a few different gps models and picks the one with the lowest deviance
        """
        models_to_try_dict = {
            "normal_gps_model": self._create_normal_gps_function(),
            "lognormal_gps_model": self._create_lognormal_gps_function(),
            "gamma_gps_model": self._create_gamma_gps_function(),
        }

        model_comparison_dict = {}

        for key, value in models_to_try_dict.items():
            model_comparison_dict[key] = value[1]

        # Return model with lowest deviance
        best_model = min(model_comparison_dict, key=model_comparison_dict.get)

        return (
            best_model,
            models_to_try_dict[best_model][0],
            models_to_try_dict[best_model][1],
        )



"""
Defines the Targetted Maximum likelihood Estimation (TMLE) model class
"""
from pprint import pprint

import numpy as np
import pandas as pd
from pandas.api.types import is_float_dtype, is_numeric_dtype
from pygam import LinearGAM, s
from scipy.interpolate import interp1d
from sklearn.neighbors import KernelDensity
from sklearn.ensemble import GradientBoostingRegressor
from statsmodels.nonparametric.kernel_regression import KernelReg

from causal_curve.core import Core


class TMLE_Core(Core):
    """
    Constructs a causal dose response curve via a modified version of Targetted
    Maximum Likelihood Estimation (TMLE) across a grid of the treatment values.
    Gradient boosting is used for prediction of the Q model and G models, simple
    kernel regression is used processing those model results, and a generalized
    additive model is used in the final step to contruct the final curve.
    Assumes continuous treatment and outcome variable.

    WARNING:

    -The treatment values should be roughly normally-distributed for this tool
    to work. Otherwise you may encounter internal math errors.

    -This algorithm assumes you've already performed the necessary transformations to
    categorical covariates (i.e. these variables are already one-hot encoded and
    one of the categories is excluded for each set of dummy variables).

    -Please take care to ensure that the "ignorability" assumption is met (i.e.
    all strong confounders are captured in your covariates and there is no
    informative censoring), otherwise your results will be biased, sometimes strongly so.

    Parameters
    ----------

    treatment_grid_num: int, optional (default = 100)
        Takes the treatment, and creates a quantile-based grid across its values.
        For instance, if the number 6 is selected, this means the algorithm will only take
        the 6 treatment variable values at approximately the 0, 20, 40, 60, 80, and 100th
        percentiles to estimate the causal dose response curve.
        Higher value here means the final curve will be more finely estimated,
        but also increases computation time. Default is usually a reasonable number.

    lower_grid_constraint:  float, optional(default = 0.01)
        This adds an optional constraint of the lower side of the treatment grid.
        Sometimes data near the minimum values of the treatment are few in number
        and thus generate unstable estimates. By default, this clips the bottom 1 percentile
        or lower of treatment values. This can be as low as 0, indicating there is no
        lower limit to how much treatment data is considered.

    upper_grid_constraint: float, optional (default = 0.99)
        See above parameter. Just like above, but this is an upper constraint.
        By default, this clips the top 99th percentile or higher of treatment values.
        This can be as high as 1.0, indicating there is no upper limit to how much
        treatment data is considered.

    n_estimators: int, optional (default = 200)
        Optional argument to set the number of learners to use when sklearn
        creates TMLE's Q and G models.

    learning_rate: float, optional (default = 0.01)
        Optional argument to set the sklearn's learning rate for TMLE's Q and G models.

    max_depth: int, optional (default = 3)
        Optional argument to set sklearn's maximum depth when creating TMLE's Q and G models.

    bandwidth: float, optional (default = 0.5)
        Optional argument to set the bandwidth parameter of the internal
        kernel density estimation and kernel regression methods.

    random_seed: int, optional (default = None)
        Sets the random seed.

    verbose: bool, optional (default = False)
        Determines whether the user will get verbose status updates.


    Attributes
    ----------



    Methods
    ----------
    fit: (self, T, X, y)
        Fits the causal dose-response model

    calculate_CDRC: (self, ci, CDRC_grid_num)
        Calculates the CDRC (and confidence interval) from TMLE estimation


    Examples
    --------



    References
    ----------

    Kennedy EH, Ma Z, McHugh MD, Small DS. Nonparametric methods for doubly robust estimation
    of continuous treatment effects. Journal of the Royal Statistical Society, Series B. 79(4), 2017, pp.1229-1245.

    van der Laan MJ and Rubin D. Targeted maximum likelihood learning. In: The International
    Journal of Biostatistics, 2(1), 2006.

    van der Laan MJ and Gruber S. Collaborative double robust penalized targeted
    maximum likelihood estimation. In: The International Journal of Biostatistics 6(1), 2010.

    """

    def __init__(
        self,
        treatment_grid_num=100,
        lower_grid_constraint=0.01,
        upper_grid_constraint=0.99,
        n_estimators=200,
        learning_rate=0.01,
        max_depth=3,
        bandwidth=0.5,
        random_seed=None,
        verbose=False,
    ):

        self.treatment_grid_num = treatment_grid_num
        self.lower_grid_constraint = lower_grid_constraint
        self.upper_grid_constraint = upper_grid_constraint
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.max_depth = max_depth
        self.bandwidth = bandwidth
        self.random_seed = random_seed
        self.verbose = verbose

        # Validate the params
        self._validate_init_params()
        self.rand_seed_wrapper()

        self.if_verbose_print("Using the following params for TMLE model:")
        if self.verbose:
            pprint(self.get_params(), indent=4)

    def _validate_init_params(self):
        """
        Checks that the params used when instantiating TMLE model are formatted correctly
        """

        # Checks for treatment_grid_num
        if not isinstance(self.treatment_grid_num, int):
            raise TypeError(
                f"treatment_grid_num parameter must be an integer, "
                f"but found type {type(self.treatment_grid_num)}"
            )

        if (isinstance(self.treatment_grid_num, int)) and self.treatment_grid_num < 10:
            raise ValueError(
                f"treatment_grid_num parameter should be >= 10 so your final curve "
                f"has enough resolution, but found value {self.treatment_grid_num}"
            )

        if (
            isinstance(self.treatment_grid_num, int)
        ) and self.treatment_grid_num >= 1000:
            raise ValueError("treatment_grid_num parameter is too high!")

        # Checks for lower_grid_constraint
        if not isinstance(self.lower_grid_constraint, float):
            raise TypeError(
                f"lower_grid_constraint parameter must be a float, "
                f"but found type {type(self.lower_grid_constraint)}"
            )

        if (
            isinstance(self.lower_grid_constraint, float)
        ) and self.lower_grid_constraint < 0:
            raise ValueError(
                f"lower_grid_constraint parameter cannot be < 0, "
                f"but found value {self.lower_grid_constraint}"
            )

        if (
            isinstance(self.lower_grid_constraint, float)
        ) and self.lower_grid_constraint >= 1.0:
            raise ValueError(
                f"lower_grid_constraint parameter cannot >= 1.0, "
                f"but found value {self.lower_grid_constraint}"
            )

        # Checks for upper_grid_constraint
        if not isinstance(self.upper_grid_constraint, float):
            raise TypeError(
                f"upper_grid_constraint parameter must be a float, "
                f"but found type {type(self.upper_grid_constraint)}"
            )

        if (
            isinstance(self.upper_grid_constraint, float)
        ) and self.upper_grid_constraint <= 0:
            raise ValueError(
                f"upper_grid_constraint parameter cannot be <= 0, "
                f"but found value {self.upper_grid_constraint}"
            )

        if (
            isinstance(self.upper_grid_constraint, float)
        ) and self.upper_grid_constraint > 1.0:
            raise ValueError(
                f"upper_grid_constraint parameter cannot > 1.0, "
                f"but found value {self.upper_grid_constraint}"
            )

        # Checks for lower_grid_constraint isn't higher than upper_grid_constraint
        if self.lower_grid_constraint >= self.upper_grid_constraint:
            raise ValueError(
                "lower_grid_constraint should be lower than upper_grid_constraint!"
            )

        # Checks for n_estimators
        if not isinstance(self.n_estimators, int):
            raise TypeError(
                f"n_estimators parameter must be an integer, "
                f"but found type {type(self.n_estimators)}"
            )

        if (self.n_estimators < 10) or (self.n_estimators > 100000):
            raise TypeError("n_estimators parameter must be between 10 and 100000")

        # Checks for learning_rate
        if not isinstance(self.learning_rate, (int, float)):
            raise TypeError(
                f"learning_rate parameter must be an integer or float, "
                f"but found type {type(self.learning_rate)}"
            )

        if (self.learning_rate <= 0) or (self.learning_rate >= 1000):
            raise TypeError(
                "learning_rate parameter must be greater than 0 and less than 1000"
            )

        # Checks for max_depth
        if not isinstance(self.max_depth, int):
            raise TypeError(
                f"max_depth parameter must be an integer, "
                f"but found type {type(self.max_depth)}"
            )

        if self.max_depth <= 0:
            raise TypeError("max_depth parameter must be greater than 0")

        # Checks for bandwidth
        if not isinstance(self.bandwidth, (int, float)):
            raise TypeError(
                f"bandwidth parameter must be an integer or float, "
                f"but found type {type(self.bandwidth)}"
            )

        if (self.bandwidth <= 0) or (self.bandwidth >= 1000):
            raise TypeError(
                "bandwidth parameter must be greater than 0 and less than 1000"
            )

        # Checks for random_seed
        if not isinstance(self.random_seed, (int, type(None))):
            raise TypeError(
                f"random_seed parameter must be an int, but found type {type(self.random_seed)}"
            )

        if (isinstance(self.random_seed, int)) and self.random_seed < 0:
            raise ValueError("random_seed parameter must be > 0")

        # Checks for verbose
        if not isinstance(self.verbose, bool):
            raise TypeError(
                f"verbose parameter must be a boolean type, but found type {type(self.verbose)}"
            )

    def _validate_fit_data(self):
        """Verifies that T, X, and y are formatted the right way"""
        # Checks for T column
        if not is_float_dtype(self.t_data):
            raise TypeError("Treatment data must be of type float")

        # Make sure all X columns are float or int
        if isinstance(self.x_data, pd.Series):
            if not is_numeric_dtype(self.x_data):
                raise TypeError(
                    "All covariate (X) columns must be int or float type (i.e. must be numeric)"
                )

        elif isinstance(self.x_data, pd.DataFrame):
            for column in self.x_data:
                if not is_numeric_dtype(self.x_data[column]):
                    raise TypeError(
                        """All covariate (X) columns must be int or float type
                        (i.e. must be numeric)"""
                    )

        # Checks for Y column
        if not is_float_dtype(self.y_data):
            raise TypeError("Outcome data must be of type float")

    def _validate_calculate_CDRC_params(self, ci):
        """Validates the parameters given to `calculate_CDRC`"""

        if not isinstance(ci, float):
            raise TypeError(
                f"`ci` parameter must be an float, but found type {type(ci)}"
            )

        if isinstance(ci, float) and ((ci <= 0) or (ci >= 1.0)):
            raise ValueError("`ci` parameter should be between (0, 1)")

    def fit(self, T, X, y):
        """Fits the TMLE causal dose-response model. For now, this only
        accepts pandas columns. You *must* provide at least one covariate column.

        Parameters
        ----------
        T: array-like, shape (n_samples,)
            A continuous treatment variable
        X: array-like, shape (n_samples, m_features)
            Covariates, where n_samples is the number of samples
            and m_features is the number of features
        y: array-like, shape (n_samples,)
            Outcome variable

        Returns
        ----------
        self : object

        """
        self.t_data = T.reset_index(drop=True, inplace=False)
        self.x_data = X.reset_index(drop=True, inplace=False)
        self.y_data = y.reset_index(drop=True, inplace=False)

        # Validate this input data
        self._validate_fit_data()

        # Capture covariate and treatment column names
        self.treatment_col_name = self.t_data.name

        if len(self.x_data.shape) == 1:
            self.covariate_col_names = [self.x_data.name]
        else:
            self.covariate_col_names = self.x_data.columns.values.tolist()

        # Note the size of the data
        self.num_rows = len(self.t_data)

        # Produce expanded versions of the inputs
        self.if_verbose_print("Transforming data for the Q-model and G-model")
        self.grid_values, self.fully_expanded_x , self.fully_expanded_t_and_x = self._transform_inputs()

        # Fit G-model and get relevent predictions
        self.if_verbose_print("Fitting G-model and making treatment assignment predictions...")
        self.g_model_preds, self.g_model_2_preds = self._g_model()

        # Fit Q-model and get relevent predictions
        self.if_verbose_print("Fitting Q-model and making outcome predictions...")
        self.q_model_preds = self._q_model()

        # Calculating treatment assignment adjustment using G-model's predictions
        self.if_verbose_print("Calculating treatment assignment adjustment using G-model's predictions...")
        self.n_interpd_values, self.var_n_interpd_values = self._treatment_assignment_correction()

        # Adjusting outcome using Q-model's predictions
        self.if_verbose_print("Adjusting outcome using Q-model's predictions...")
        self.outcome_adjust, self.expand_outcome_adjust = self._outcome_adjustment()

        # Calculating corrected pseudo-outcome values
        self.if_verbose_print("Calculating corrected pseudo-outcome values...")
        self.pseudo_out = (self.y_data - self.outcome_adjust) / (self.n_interpd_values / self.var_n_interpd_values) + self.expand_outcome_adjust

        # Training final GAM model using pseudo-outcome values
        self.if_verbose_print("Training final GAM model using pseudo-outcome values...")
        self.final_gam = self._fit_final_gam()


    def calculate_CDRC(self, ci=0.95):
        """Using the results of the fitted model, this generates a dataframe of CDRC point estimates
        at each of the values of the treatment grid. Connecting these estimates will produce
        the overall estimated CDRC. Confidence interval is returned as well.

        Parameters
        ----------
        ci: float (default = 0.95)
            The desired confidence interval to produce. Default value is 0.95, corresponding
            to 95% confidence intervals. bounded (0, 1.0).

        Returns
        ----------
        dataframe: Pandas dataframe
            Contains treatment grid values, the CDRC point estimate at that value,
            and the associated lower and upper confidence interval bounds at that point.

        self: object

        """
        self._validate_calculate_CDRC_params(ci)

        self.if_verbose_print("""
            Generating predictions for each value of treatment grid,
            and averaging to get the CDRC..."""
        )

        # Create CDRC predictions from the trained, final GAM

        self._cdrc_preds = self._cdrc_predictions_continuous(ci)

        return pd.DataFrame(
            self._cdrc_preds, columns=["Treatment", "Causal_Dose_Response", "Lower_CI", "Upper_CI"]
        ).round(3)


    def _transform_inputs(self):
        """Takes the treatment and covariates and transforms so they can
        be used by the algo"""

        # Create treatment grid
        grid_values = np.linspace(
            start=self.t_data.min(),
            stop=self.t_data.max(),
            num=self.treatment_grid_num
        )

        # Create expanded treatment array
        expanded_t = np.array([])
        for treat_value in grid_values:
        	expanded_t = np.append(expanded_t, ([treat_value] * self.num_rows))

        # Create expanded treatment array with covariates
        expanded_t_and_x = pd.concat(
            [
                pd.DataFrame(expanded_t),
                pd.concat(
                    [self.x_data] * self.treatment_grid_num
                ).reset_index(drop = True, inplace = False),
            ],
	        axis = 1,
            ignore_index = True
        )

        expanded_t_and_x.columns = [self.treatment_col_name] + self.covariate_col_names

        fully_expanded_t_and_x = pd.concat(
        	[
        		pd.concat([self.x_data, self.t_data], axis=1),
        		expanded_t_and_x
        	],
        	axis = 0,
        	ignore_index = True
        )

        fully_expanded_x = fully_expanded_t_and_x[self.covariate_col_names]

        return grid_values, fully_expanded_x, fully_expanded_t_and_x

    def _g_model(self):
        """Produces the G-model and gets treatment assignment predictions"""

        t = self.t_data.to_numpy()
        X = self.x_data.to_numpy()

        g_model = GradientBoostingRegressor(
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            learning_rate=self.learning_rate,
            random_state=self.random_seed
        ).fit(X = X, y = t)
        g_model_preds = g_model.predict(self.fully_expanded_x)

        g_model2 = GradientBoostingRegressor(
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            learning_rate=self.learning_rate,
            random_state=self.random_seed
        ).fit(X = X, y = ((t - g_model_preds[0:self.num_rows])**2))
        g_model_2_preds = g_model2.predict(self.fully_expanded_x)

        return g_model_preds, g_model_2_preds

    def _q_model(self):
        """Produces the Q-model and gets outcome predictions using the provided treatment
        values, when the treatment is completely present and not present.
        """

        X = pd.concat([self.t_data, self.x_data], axis=1).to_numpy()
        y = self.y_data.to_numpy()

        q_model = GradientBoostingRegressor(
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            learning_rate=self.learning_rate,
            random_state=self.random_seed
        ).fit(X = X, y = y)
        q_model_preds = q_model.predict(self.fully_expanded_t_and_x)

        return q_model_preds


    def _treatment_assignment_correction(self):
        """Uses the G-model and its predictions to adjust treatment assignment
        """

        t_standard = (
            (self.fully_expanded_t_and_x[self.treatment_col_name] - self.g_model_preds) / np.sqrt(self.g_model_2_preds)
        )

        interpd_values = interp1d(
            self.one_dim_estimate_density(t_standard.values)[0],
            self.one_dim_estimate_density(t_standard.values[0:self.num_rows])[1],
            kind='linear'
        )(t_standard) / np.sqrt(self.g_model_2_preds)

        n_interpd_values = interpd_values[0:self.num_rows]

        temp_interpd = interpd_values[self.num_rows:]

        zeros_mat = np.zeros((self.num_rows, self.treatment_grid_num))

        for i in range(0, self.treatment_grid_num):
        	lower = i * self.num_rows
        	upper = i * self.num_rows + self.num_rows
        	zeros_mat[:,i] = temp_interpd[lower:upper]

        var_n_interpd_values = self.pred_from_loess(
            train_x = self.grid_values,
            train_y = zeros_mat.mean(axis = 0),
            x_to_pred = self.t_data
        )

        return n_interpd_values, var_n_interpd_values


    def _outcome_adjustment(self):
        """Uses the Q-model and its predictions to adjust the outcome
        """

        outcome_adjust = self.q_model_preds[0:self.num_rows]

        temp_outcome_adjust = self.q_model_preds[self.num_rows:]

        zero_mat = np.zeros((self.num_rows, self.treatment_grid_num))
        for i in range(0, self.treatment_grid_num):
        	lower = i * self.num_rows
        	upper = i * self.num_rows + self.num_rows
        	zero_mat[:,i] = temp_outcome_adjust[lower:upper]

        expand_outcome_adjust = self.pred_from_loess(
            train_x = self.grid_values,
            train_y = zero_mat.mean(axis = 0),
            x_to_pred = self.t_data
        )

        return outcome_adjust, expand_outcome_adjust

    def _fit_final_gam(self):
        """We now regress the original treatment values against the pseudo-outcome values
        """

        return LinearGAM(
        	s(0, n_splines=30, spline_order=3),
            max_iter=500,
            lam=self.bandwidth
        ).fit(self.t_data, y = self.pseudo_out)

    def one_dim_estimate_density(self, series):
    	"""
    	Takes in a numpy array, returns grid values for KDE and predicted probabilities
    	"""
    	series_grid = np.linspace(
            start=series.min(),
            stop=series.max(),
            num=self.num_rows
        )

    	kde = KernelDensity(
            kernel='gaussian',
            bandwidth=self.bandwidth
        ).fit(series.reshape(-1, 1))

    	return series_grid, np.exp(kde.score_samples(series_grid.reshape(-1, 1)))

    def pred_from_loess(self, train_x, train_y, x_to_pred):
    	"""
    	Trains simple loess regression and returns predictions
    	"""
    	kr_model = KernelReg(
            endog = train_y,
            exog = train_x,
            var_type = 'c',
            bw = [self.bandwidth]
        )

    	return kr_model.fit(x_to_pred)[0]



"""
Defines the Targetted Maximum likelihood Estimation (TMLE) regressor model class
"""

import numpy as np

from causal_curve.tmle_core import TMLE_Core


class TMLE_Regressor(TMLE_Core):
    """
    A TMLE tool that handles continuous outcomes. Inherits the TMLE_core
    base class. See that base class code its docstring for more details.
    """

    def _cdrc_predictions_continuous(self, ci):
        """Returns the predictions of CDRC for each value of the treatment grid. Essentially,
        we're making predictions using the original treatment against the pseudo-outcome.
        To be used when the outcome of interest is continuous.
        """

        # To keep track of cdrc predictions, we create an empty 2d array of shape
        # (treatment_grid_num, 4). The last dimension is of length 4 because
        # we are going to keep track of the treatment, point estimate of the prediction, as well as
        # the lower and upper bounds of the prediction interval
        cdrc_preds = np.zeros((self.treatment_grid_num, 4), dtype=float)

        # Loop through each of the grid values, get point estimate and get estimate interval
        for i in range(0, self.treatment_grid_num):
            temp_T = self.grid_values[i]
            temp_cdrc_preds = self.final_gam.predict(temp_T)
            temp_cdrc_interval = self.final_gam.confidence_intervals(temp_T, width=ci)
            temp_cdrc_lower_bound = temp_cdrc_interval[:, 0]
            temp_cdrc_upper_bound = temp_cdrc_interval[:, 1]
            cdrc_preds[i, 0] = temp_T
            cdrc_preds[i, 1] = temp_cdrc_preds
            cdrc_preds[i, 2] = temp_cdrc_lower_bound
            cdrc_preds[i, 3] = temp_cdrc_upper_bound

        return cdrc_preds

    def point_estimate(self, T):
        """Calculates point estimate within the CDRC given treatment values.
        Can only be used when outcome is continuous. Can be estimate for a single
        data point or can be run in batch for many observations. Extrapolation will produce
        untrustworthy results; the provided treatment should be within
        the range of the training data.

        Parameters
        ----------
        T: Numpy array, shape (n_samples,)
            A continuous treatment variable.

        Returns
        ----------
        array: Numpy array
            Contains a set of CDRC point estimates

        """
        return np.apply_along_axis(self._create_point_estimate, 0, T.reshape(1, -1))

    def _create_point_estimate(self, T):
        """Takes a single treatment value and produces a single point estimate
        in the case of a continuous outcome.
        """
        return self.final_gam.predict(np.array([T]).reshape(1, -1))

    def point_estimate_interval(self, T, ci=0.95):
        """Calculates the prediction confidence interval associated with a point estimate
        within the CDRC given treatment values. Can only be used
        when outcome is continuous. Can be estimate for a single data point
        or can be run in batch for many observations. Extrapolation will produce
        untrustworthy results; the provided treatment should be within
        the range of the training data.

        Parameters
        ----------
        T: Numpy array, shape (n_samples,)
            A continuous treatment variable.
        ci: float (default = 0.95)
            The desired confidence interval to produce. Default value is 0.95, corresponding
            to 95% confidence intervals. bounded (0, 1.0).

        Returns
        ----------
        array: Numpy array
            Contains a set of CDRC prediction intervals ([lower bound, higher bound])

        """
        return np.apply_along_axis(
            self._create_point_estimate_interval, 0, T.reshape(1, -1), width=ci
        ).T.reshape(-1, 2)

    def _create_point_estimate_interval(self, T, width):
        """Takes a single treatment value and produces confidence interval
        associated with a point estimate in the case of a continuous outcome.
        """
        return self.final_gam.prediction_intervals(
            np.array([T]).reshape(1, -1), width=width
        )



"""
Defines the Generalized Prospensity Score (GPS) classifier model class
"""

import numpy as np
from scipy.special import logit

from causal_curve.gps_core import GPS_Core


class GPS_Classifier(GPS_Core):
    """
    A GPS tool that handles binary outcomes. Inherits the GPS_core
    base class. See that base class code its docstring for more details.
    """

    def __init__(
        self,
        gps_family=None,
        treatment_grid_num=100,
        lower_grid_constraint=0.01,
        upper_grid_constraint=0.99,
        spline_order=3,
        n_splines=30,
        lambda_=0.5,
        max_iter=100,
        random_seed=None,
        verbose=False,
    ):
        GPS_Core.__init__(
            self,
            gps_family=None,
            treatment_grid_num=100,
            lower_grid_constraint=0.01,
            upper_grid_constraint=0.99,
            spline_order=3,
            n_splines=30,
            lambda_=0.5,
            max_iter=100,
            random_seed=None,
            verbose=False,
        )

    def _cdrc_predictions_binary(self, ci):
        """Returns the predictions of CDRC for each value of the treatment grid. Essentially,
        we're making predictions using the original treatment and gps_at_grid.
        To be used when the outcome of interest is binary.
        """
        # To keep track of cdrc predictions, we create an empty 2d array of shape
        # (n_samples, treatment_grid_num, 2). The last dimension is of length 2 because
        # we are going to keep track of the point estimate (log-odds) of the prediction, as well as
        # the standard error of the prediction interval (again, this is for the log odds)
        cdrc_preds = np.zeros((len(self.T), self.treatment_grid_num, 2), dtype=float)

        # Loop through each of the grid values, predict point estimate and get prediction interval
        for i in range(0, self.treatment_grid_num):

            temp_T = np.repeat(self.grid_values[i], repeats=len(self.T))
            temp_gps = self.gps_at_grid[:, i]

            temp_cdrc_preds = logit(
                self.gam_results.predict_proba(np.column_stack((temp_T, temp_gps)))
            )

            temp_cdrc_interval = logit(
                self.gam_results.confidence_intervals(
                    np.column_stack((temp_T, temp_gps)), width=ci
                )
            )

            standard_error = (
                temp_cdrc_interval[:, 1] - temp_cdrc_preds
            ) / self.calculate_z_score(ci)

            cdrc_preds[:, i, 0] = temp_cdrc_preds
            cdrc_preds[:, i, 1] = standard_error

        return np.round(cdrc_preds, 3)

    def estimate_log_odds(self, T):
        """Calculates the estimated log odds of the highest integer class. Can
        only be used when the outcome is binary. Can be estimate for a single
        data point or can be run in batch for many observations. Extrapolation will produce
        untrustworthy results; the provided treatment should be within
        the range of the training data.

        Parameters
        ----------
        T: Numpy array, shape (n_samples,)
            A continuous treatment variable.

        Returns
        ----------
        array: Numpy array
            Contains a set of log odds
        """
        if self.outcome_type != "binary":
            raise TypeError("Your outcome must be binary to use this function!")

        return np.apply_along_axis(self._create_log_odds, 0, T.reshape(1, -1))

    def _create_log_odds(self, T):
        """Take a single treatment value and produces the log odds of the higher
        integer class, in the case of a binary outcome.
        """
        return logit(
            self.gam_results.predict_proba(
                np.array([T, self.gps_function(T).mean()]).reshape(1, -1)
            )
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="causal-curve",
    version="1.0.1",
    author="Roni Kobrosly",
    author_email="roni.kobrosly@gmail.com",
    description="A python library with tools to perform causal inference using \
        observational data when the treatment of interest is continuous.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ronikobrosly/causal-curve",
    packages=setuptools.find_packages(include=['causal_curve']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'black',
        'coverage',
        'future',
        'joblib',
        'numpy',
        'numpydoc',
        'pandas',
        'patsy',
        'progressbar2',
        'pygam',
        'pytest',
        'python-dateutil',
        'python-utils',
        'pytz',
        'scikit-learn',
        'scipy',
        'six',
        'sphinx_rtd_theme',
        'statsmodels'
    ]
)
